from django.db import models
from django.utils.translation import gettext_lazy as _

from study_registration.studies.models import Study
from study_registration.utils.fields import DateField


class DeployTarget(models.Model):
    name = models.CharField(_('Name'), max_length=256, unique=True)
    key = models.SlugField(_('key'), max_length=32, unique=True)
    gitlab_deploy_key = models.PositiveIntegerField(
        _('GitLab deploy key'), blank=True, null=True
    )
    url_template = models.CharField(
        _('Target URL template'), max_length=256, blank=True
    )

    class Meta:
        ordering = ['name']

    def __str__(self):
        return self.name


class BaseDeployment(models.Model):
    class Meta:
        abstract = True

    created_at = models.DateField(_('Created at'), auto_now_add=True)
    git_ref = models.CharField(_('Branch'), max_length=64, default='main')
    git_file = models.CharField(
        _('Name of the file within the repository that starts your experiment'),
        help_text=_(
            'A desktop shortcut to this file will be created on the '
            'laboratory computers.'
        ),
        max_length=256,
    )
    commit_sha = models.CharField(_('Commit (version)'), max_length=50)
    startdate = DateField(_('Start of deployment'))
    enddate = DateField(
        _('End of deployment'),
        help_text=_(
            'Your files and collected data will automatically be deleted '
            'from the deployment target after that date.'
        ),
    )


class Deployment(BaseDeployment):
    NOT_RUNNIG = 0
    RUNNING = 10
    FINISHED = 20

    study = models.ForeignKey(Study, on_delete=models.CASCADE, verbose_name=_('Study'))
    legacy_study_id = models.CharField(_('Legacy study ID'), max_length=254, blank=True)

    git_repo = models.CharField(_('Repository URL'), max_length=256)
    gitlab_id = models.IntegerField(
        _('GitLab project ID'),
        help_text=_(
            'The project ID is displayed right below the project title on GitLab.'
        ),
    )
    target = models.ForeignKey(
        DeployTarget, verbose_name=_('Deploy target'), on_delete=models.PROTECT
    )
    targeturl = models.CharField(_('Target URL'), max_length=300, blank=True)
    status = models.IntegerField(_('Status'), choices=[
        (NOT_RUNNIG, _('Not running')),
        (RUNNING, _('Running')),
        (FINISHED, _('Finished')),
    ], default=NOT_RUNNIG)

    class Meta:
        permissions = [
            ('approve', _('Can approve')),
        ]

    def __str__(self):
        return '{} {} {}'.format(
            self.git_repo.split('/')[-1],
            self.git_ref,
            self.git_file,
        )

    @property
    def generated_branch(self):
        return f'asr_deploy_to_{self.target.key}'

    def create_revision(self):
        self.revisions.create(
            deployment=self,
            git_ref=self.git_ref,
            git_file=self.git_file,
            commit_sha=self.commit_sha,
            startdate=self.startdate,
            enddate=self.enddate,
        )


class DeploymentRevision(BaseDeployment):
    deployment = models.ForeignKey(
        Deployment, on_delete=models.CASCADE, related_name='revisions'
    )
