from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from study_registration.utils.fields import DateField


class Department(models.Model):
    name = models.CharField(_('Name'), max_length=254, unique=True)
    ldap_group = models.CharField(_('LDAP group'), max_length=254, blank=True)
    enable_deployments = models.BooleanField(_('Enable deployments'), default=False)
    enable_budgets = models.BooleanField(_('Enable budgets'), default=False)

    class Meta:
        ordering = ['name']

    def __str__(self):
        return self.name


class Project(models.Model):
    created_at = models.DateField(_('Created at'), auto_now_add=True)
    title = models.CharField(_('Title'), max_length=255, unique=True)

    class Meta:
        ordering = ['title']

    def __str__(self):
        return self.title


class EthicsBoard(models.Model):
    name = models.CharField(_('Name'), max_length=254, unique=True)

    class Meta:
        ordering = ['name']

    def __str__(self):
        return self.name


class Study(models.Model):
    created_at = models.DateField(_('Created at'), auto_now_add=True)
    updated_at = models.DateField(_('Updated at'), auto_now=True)
    imported_from = models.CharField(_('Imported from'), max_length=254, blank=True)
    owners = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        verbose_name=_(
            'Users who can edit this study profile and access the private folder'
        ),
    )

    title = models.CharField(_('Title'), max_length=254)
    short_title = models.SlugField(
        _('Short title'),
        help_text=_(
            'The short title will be used to refer to this study within '
            'the institute context, e.g., to create a physical folder '
            'on the server. Therefore, only use letters, numbers, '
            'hyphens, or underscores.'
        ),
        max_length=254,
    )
    aliases = models.CharField(
        _('Other titles or acronyms'),
        help_text=_(
            'Please compile a complete list of additional titles that '
            'may refer to your study, including e.g. a German title '
            'used in communication with participants'
        ),
        max_length=254,
        blank=True,
    )
    end = DateField(
        _('Expected end of study'),
        help_text=_(
            'When analyses are completed, manuscripts are submitted, and the data is '
            'no longer in active use, it should be moved to an archive in order to '
            'free up storage capacity. Entering this date allows the Rasearch Data '
            'Management Team to contact you to coordinate the archiving process.'
        ),
        blank=True,
        null=True,
    )
    description = models.TextField(
        _('Description'),
        help_text=_(
            'This text should allow other researchers to quickly grasp whether this '
            'study and the related data could be of interest to them.'
        ),
        blank=True,
    )
    department = models.ForeignKey(
        Department, verbose_name=_('Department'), on_delete=models.CASCADE
    )
    project = models.ForeignKey(
        Project,
        verbose_name=_('Project'),
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )

    contact_person = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_('Contact person'),
        help_text=_(
            'The contact person must be reachable. Before the current contact person '
            'leaves the institute, they must pass on the study to someone else.'
        ),
        on_delete=models.PROTECT,
        limit_choices_to={
            'is_active': True,
        },
        related_name='+',
    )
    affiliated_scientists = models.CharField(
        _('Affiliated Scientists'),
        max_length=254,
        help_text=_(
            'You can use the same format as you would use for the authors of a paper.'
        ),
    )

    ethics_board = models.ForeignKey(
        EthicsBoard,
        verbose_name=_('Ethics board'),
        help_text=_(
            'Please remember to save the application and approval to the study folder.'
        ),
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )

    storage_location = models.CharField(
        _('Storage location'), max_length=255, blank=True, null=True, unique=True
    )
    storage_archived = models.DateField(_('Storage archived'), blank=True, null=True)
    storage_managed = models.BooleanField(_('Storage is managed'), default=False)

    class Meta:
        ordering = ['-pk']

    def __str__(self):
        return '{}-{}'.format(self.pk, self.short_title)

    def get_absolute_url(self):
        return reverse(
            'studies:detail', kwargs={'pk': self.pk, 'slug': self.short_title}
        )


class Keyword(models.Model):
    study = models.ForeignKey(Study, on_delete=models.CASCADE)
    name = models.CharField(_('Name'), max_length=128)

    class Meta:
        unique_together = [['study', 'name']]

    def __str__(self):
        return self.name


class Link(models.Model):
    WEBSITE = 1
    PREREG = 2
    DATA = 3
    CODE = 4

    study = models.ForeignKey(Study, on_delete=models.CASCADE)
    url = models.URLField(_('URL'))
    type = models.SmallIntegerField(_('Type'), choices=[
        (WEBSITE, _('Website')),
        (PREREG, _('Pre-Registration')),
        (DATA, _('Public data repository')),
        (CODE, _('Public code repository')),
    ])

    class Meta:
        ordering = ['type', 'pk']
