import datetime
import re
from urllib.parse import urlencode

from django import forms
from django.conf import settings
from django.contrib import messages
from django.contrib.auth.mixins import LoginRequiredMixin
from django.core.mail import send_mail
from django.db import models
from django.db.models.functions import Coalesce
from django.shortcuts import get_object_or_404
from django.shortcuts import redirect
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import CreateView
from django.views.generic import DeleteView
from django.views.generic import DetailView
from django.views.generic import ListView
from django.views.generic import RedirectView
from django.views.generic import TemplateView
from django.views.generic import UpdateView

from study_registration.datasets.models import DataSet
from study_registration.utils import smb
from study_registration.utils.search import setsearchquery

from . import storage
from .forms import StudyCreateForm
from .forms import StudyUpdateForm
from .models import Department
from .models import EthicsBoard
from .models import Keyword
from .models import Link
from .models import Project
from .models import Study


def split_with_quotes(s):
    part = ''
    quoted = False
    for c in s:
        if quoted:
            if c == '"':
                quoted = False
            else:
                part += c
        else:
            if c == ' ':
                if part:
                    yield part
                    part = ''
            elif c == '"':
                quoted = True
            else:
                part += c
    if part:
        yield part


def make_score(value, **kwargs):
    return models.Value(value) * models.Count('pk', filter=models.Q(**kwargs))


def make_keyword_score(value, **kwargs):
    return models.Value(value) * Coalesce((
        Keyword.objects
        .filter(study=models.OuterRef('pk'), **kwargs)
        .order_by()
        .values('study')
        .annotate(count=models.Count('study'))
        .values('count')
    ), 0)


class OwnerRequiredMixin(LoginRequiredMixin):
    def get_queryset(self):
        if self.request.user.is_superuser:
            return Study.objects.all()
        else:
            return self.request.user.study_set.all()


class StudyListView(ListView):
    model = Study
    paginate_by = 20

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        def choices(model, field):
            return [
                (name, name) for name in model.objects.values_list(field, flat=True)
            ]

        reuse = DataSet._meta.get_field('data_reuse')

        context['filters'] = [
            ('department', _('Department'), choices(Department, 'name')),
            ('project', _('Project'), choices(Project, 'title')),
            ('ethicsboard', _('Ethics Board'), choices(EthicsBoard, 'name')),
            ('reuse', reuse.verbose_name, reuse.choices[1:]),
            ('open', _('Open Science'), [
                ('prereg', _('Pre-Registration')),
                ('data', _('Public data repository')),
                ('code', _('Public code repository')),
            ]),
        ]
        return context

    def apply_score(self, qs, q):
        score = models.Value(1)
        for part in split_with_quotes(q):
            if ':' in part:
                key, value = part.split(':', 1)
                if key == 'contact':
                    qs = qs.filter(contact_person__username=value)
                    continue
                elif key == 'owner':
                    qs = qs.filter(owners__username=value)
                    continue
                elif key == 'department':
                    qs = qs.filter(department__name=value)
                    continue
                elif key == 'project':
                    qs = qs.filter(project__title=value)
                    continue
                elif key == 'keyword':
                    qs = qs.filter(keyword__name=value)
                    continue
                elif key == 'doi':
                    qs = qs.filter(publication__doi=value)
                    continue
                elif key == 'reuse':
                    qs = qs.filter(dataset__data_reuse__gte=value)
                    continue
                elif key == 'ethicsboard':
                    qs = qs.filter(ethics_board__name=value)
                    continue
                elif key == 'open' and value == 'prereg':
                    qs = qs.filter(has_preregistration=True)
                    continue
                elif key == 'open' and value == 'data':
                    qs = qs.filter(has_open_data=True)
                    continue
                elif key == 'open' and value == 'code':
                    qs = qs.filter(has_open_materials=True)
                    continue

            score *= sum([
                make_score(100, title__iexact=part),
                make_score(100, short_title__iexact=part),
                make_score(100, publication__doi=part),
                make_score(50, title__icontains=part),
                make_score(20, contact_person__username__icontains=part),
                make_score(10, affiliated_scientists__icontains=part),
                make_score(5, aliases__icontains=part),
                make_score(5, department__name__iexact=part),
                make_keyword_score(5, name__iexact=part),
                make_keyword_score(2, name__icontains=part),
                make_score(1, description__icontains=part),
                make_score(1, contact_person__first_name__icontains=part),
            ])

        return qs.annotate(score=score)

    def get_queryset(self):
        qs = (
            super().get_queryset()
            .select_related('contact_person')
            .prefetch_related('keyword_set')
        )

        qs = qs.annotate(
            has_preregistration=models.Exists(
                Link.objects.filter(study=models.OuterRef('pk'), type=Link.PREREG)
            ),
            has_open_data=models.Exists(
                Link.objects.filter(study=models.OuterRef('pk'), type=Link.DATA)
            ),
            has_open_materials=models.Exists(
                Link.objects.filter(study=models.OuterRef('pk'), type=Link.CODE)
            ),
        )

        q = self.request.GET.get('q')
        if q:
            qs = self.apply_score(qs, q)
            qs = qs.filter(score__gt=0).order_by('-score', '-created_at')
        else:
            qs = qs.order_by('-created_at')

        return qs.distinct()

    def get(self, request, **kwargs):
        # if q is an exact study ID, redirect there directly
        q = request.GET.get('q', '')
        m = re.match('([0-9]+)-([-a-zA-Z0-9_]+)', q)
        if m:
            study = (
                super().get_queryset()
                .filter(pk=int(m[1], 10), short_title=m[2])
                .first()
            )
            if study:
                return redirect(study.get_absolute_url())

        # if any other field is provided, encode it into q and redirect
        modified = False
        query = {}
        for key, value in request.GET.items():
            if key == 'page':
                query[key] = value
            elif key != 'q':
                q = setsearchquery(q, key, value)
                modified = True
        if modified:
            query['q'] = q
            return redirect(request.path + '?' + urlencode(query))

        return super().get(request, **kwargs)


class MyStudiesView(LoginRequiredMixin, RedirectView):
    def get_redirect_url(self):
        return '{}?{}'.format(reverse('studies:list'), urlencode({
            'q': 'owner:' + self.request.user.username,
        }))


class StudyDetailView(DetailView):
    model = Study
    slug_field = 'short_title'
    query_pk_and_slug = True

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['is_owner'] = (
            self.request.user.is_superuser
            or self.object.owners.filter(pk=self.request.user.pk).exists()
        )
        context['last_publication_year'] = self.object.publication_set.aggregate(
            year=models.Max('year')
        )['year']
        if context['last_publication_year']:
            context['publication_to_be_stored_until'] = (
                context['last_publication_year'] + 10
            )
        return context


class StudyRedirectView(RedirectView):
    def get_redirect_url(self, pk):
        study = get_object_or_404(Study, pk=pk)
        return study.get_absolute_url()


class LinkFormsetMixin:
    def get_formset(self):
        formset_class = forms.modelformset_factory(
            Link, fields=['url', 'type'], extra=1, can_delete=True
        )
        if self.object:
            qs = self.object.link_set.all()
        else:
            qs = Link.objects.none()
        return formset_class(queryset=qs, data=self.get_form_kwargs().get('data'))

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        if 'formset' not in context:
            context['formset'] = self.get_formset()
        return context

    def form_valid(self, form, formset):
        response = super().form_valid(form)
        for f in formset.extra_forms:
            f.instance.study = self.object
        formset.save()
        return response


class StudyCreateView(OwnerRequiredMixin, LinkFormsetMixin, CreateView):
    model = Study
    form_class = StudyCreateForm

    def send_notification(self):
        study = str(self.object)
        user = self.request.user.get_full_name()
        link = self.request.build_absolute_uri(self.object.get_absolute_url())

        subject = settings.EMAIL_SUBJECT_PREFIX + f'Study registered: {study}'
        body = f'{user} registered a new study:\n\n{link}'
        to = settings.RDM_NOTIFICATIONS_TO

        return send_mail(subject, body, None, to)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs['initial'] = {
            'contact_person': self.request.user,
            'department': self.request.user.default_department,
            'owners': [self.request.user],
        }
        return kwargs

    def post(self, request, *args, **kwargs):
        self.object = None
        form = self.get_form()
        formset = self.get_formset()
        if form.is_valid() and formset.is_valid():
            return self.form_valid(form, formset)
        else:
            return self.form_invalid(form, formset)

    def form_valid(self, form, *args):
        response = super().form_valid(form, *args)
        form.instance.owners.add(self.request.user)
        self.send_notification()
        return response

    def get_success_url(self):
        return reverse(
            'studies:storage', args=[self.object.pk, self.object.short_title]
        )


class StudyUpdateView(OwnerRequiredMixin, LinkFormsetMixin, UpdateView):
    model = Study
    form_class = StudyUpdateForm
    slug_field = 'short_title'
    query_pk_and_slug = True

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        form = self.get_form()
        formset = self.get_formset()
        if form.is_valid() and formset.is_valid():
            return self.form_valid(form, formset)
        else:
            return self.form_invalid(form, formset)

    def form_invalid(self, form, formset):
        return self.render_to_response(
            self.get_context_data(form=form, formset=formset)
        )

    def form_valid(self, form, *args):
        response = super().form_valid(form, *args)
        if form.instance.storage_managed:
            storage.sync(form.instance)
        return response

    def get_success_url(self):
        return reverse('studies:detail', args=[self.object.pk, self.object.short_title])


class StudyDeleteView(OwnerRequiredMixin, DeleteView):
    model = Study
    slug_field = 'short_title'
    query_pk_and_slug = True

    def send_notification(self):
        study = str(self.object)
        user = self.request.user.get_full_name()
        contact = self.object.contact_person
        storage_location = self.object.storage_location

        subject = settings.EMAIL_SUBJECT_PREFIX + f'Study deleted: {study}'
        body = (
            f'{user} deleted the study {study}:\n\n'
            f'contact person: {contact.get_full_name()} <{contact.email}>\n'
            f'storage location: {storage_location}\n\n'
            'Please check whether the storage folder should be deleted, too!'
        )
        to = settings.RDM_NOTIFICATIONS_TO

        return send_mail(subject, body, None, to)

    def get_success_url(self):
        return reverse('studies:list')

    def form_valid(self, form):
        self.send_notification()
        return super().form_valid(form)


class StorageCreateView(OwnerRequiredMixin, DetailView):
    model = Study
    slug_field = 'short_title'
    query_pk_and_slug = True
    template_name = 'studies/storage_create.html'

    def post(self, request, **kwargs):
        study = self.get_object()
        url = storage.sync(study)
        if url:
            study.storage_location = url
            study.storage_managed = True
            study.save()

            display = smb.url_or_unc(self.request, url)
            messages.success(
                self.request, _('A study folder has been created at {}').format(display)
            )
        return redirect(study.get_absolute_url())


class CalendarView(TemplateView):
    template_name = 'studies/calendar.html'

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        today = datetime.date.today()
        year = self.kwargs.get('year', today.year)
        month = self.kwargs.get('month', today.month)

        items = []

        for study in Study.objects.filter(
            created_at__year=year, created_at__month=month
        ):
            items.append((study.created_at, study, _('registration'), 'primary'))

        for study in Study.objects.filter(end__year=year, end__month=month):
            items.append((study.end, study, _('end'), 'danger'))

        for study in Study.objects.annotate(
            expected_end=models.ExpressionWrapper(
                models.F('created_at') + datetime.timedelta(days=365),
                output_field=models.DateField(),
            ),
        ).filter(end=None, expected_end__year=year, expected_end__month=month):
            items.append(
                (study.expected_end, study, _('1 year after registration'), 'danger')
            )

        for d in DataSet.objects.filter(
            collection_start__year=year,
            collection_start__month=month,
        ).select_related('study'):
            items.append(
                (d.collection_start, d.study, _('collection start'), 'success')
            )

        for d in DataSet.objects.filter(
            collection_end__year=year,
            collection_end__month=month,
        ).select_related('study'):
            items.append((d.collection_end, d.study, _('collection end'), 'warning'))

        context['year'] = year
        context['month'] = month
        context['items'] = sorted(items, key=lambda x: x[0])

        return context


class StudyMaintenanceView(ListView):
    template_name = 'studies/study_maintenance.html'
    model = Study
    paginate_by = 20

    def get_queryset(self):
        _filter = self.request.GET.get('filter', 'end')
        qs = (
            super().get_queryset()
            .select_related('contact_person')
            .prefetch_related('keyword_set')
            .order_by('created_at')
        )

        if _filter == 'archive':
            today = datetime.date.today()
            return (
                qs.filter(end__lt=today, storage_archived=None)
                .exclude(storage_location=None)
            )
        elif _filter == 'inactive':
            return qs.filter(contact_person__is_active=False)
        else:
            return qs.filter(end=None)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['filters'] = [
            ('end', _('No end date')),
            ('archive', _('Can be archived')),
            ('inactive', _('Inactive contact')),
        ]
        return context
