import datetime
import html
import sqlite3

from django.core.management.base import BaseCommand
from django.utils.text import slugify
from freezegun import freeze_time

from study_registration.deployments.models import Deployment
from study_registration.deployments.models import DeployTarget
from study_registration.studies.models import Department
from study_registration.studies.models import Study
from study_registration.users.models import User
from study_registration.utils import smb

DEPLOY_TARGETS = {
    0: '🐀 behavioural laboratory',
    1: '👀 eye tracking laboratory',
    2: '🧠 eeg laboratory',
    3: '🌐 internal virtual web laboratory',
    4: '🌍 public accessible virtual web laboratory',
    5: '🔧 test pc laboratory',
    6: '🔧 test virtual web laboratory',
}

DEPLOY_STATUS = {
    0: Deployment.NOT_RUNNIG,
    10: Deployment.RUNNING,
    20: Deployment.FINISHED,
}

DEFAULT_DATE = datetime.date(1963, 1, 1)


def parse_actualcost(s):
    if not s:
        return None
    try:
        return float(s.replace(',', '.'))
    except ValueError:
        return None


def strip_tags(s):
    return html.unescape(
        s
        .replace('<br>', '\n')
        .replace('<br />', '\n')
        .replace('&nbsp;', ' ')
        .replace('\\n', ' ')
        .replace('\\r', '')
        .replace('<p>', '\n\n')
        .replace('</p>', '\n\n')
        .replace('<h1>', '\n\n## ')
        .replace('</h1>', '\n\n')
        .replace('<h2>', '\n\n## ')
        .replace('</h2>', '\n\n')
        .replace('<strong>', '')
        .replace('</strong>', '')
    )


def import_budgets(db, study, study_row):
    approved_by = None
    if study_row['g3approvalby']:
        approved_by, _ = User.objects.get_or_create(username=study_row['g3approvalby'])

    rows = db.execute(
        'select content from budgetapproval where studyid=?',
        [study_row['id']],
    ).fetchall()

    if rows:
        for row in rows:
            study.budget_set.create(
                expected=study_row['expectedcost'],
                actual=parse_actualcost(study_row['actualcost']),
                motivation=strip_tags(row['content']),
                approved_by=approved_by,
            )
    else:
        study.budget_set.create(
            expected=study_row['expectedcost'],
            actual=parse_actualcost(study_row['actualcost']),
            approved_by=approved_by,
        )


def import_deployments(db, deploy_targets, study, study_row):
    # FIXME: should we import all or just the lastest one?
    all_rows = db.execute(
        'select '
        'gitlab_id, '
        'repository, '
        'generated_branch, '  # unused
        'commit_sha, '
        'branch, '
        'deploytarget, '
        'targeturl, '
        'status, '
        'startdate, '
        'enddate, '
        'created_at, '
        'start_script '
        'from runningstudy '
        'where studyid=?',
        [study_row['id']],
    ).fetchall()

    for target_id, target in deploy_targets.items():
        rows = [row for row in all_rows if row['deploytarget'] == target_id]

        if not rows:
            continue

        with freeze_time(rows[0]['created_at']):
            deployment = study.deployment_set.create(
                git_repo=rows[-1]['repository'],
                gitlab_id=rows[-1]['gitlab_id'],
                target=target,
                targeturl=(rows[-1]['targeturl'] or '').replace(r'\\', '\\'),
                status=DEPLOY_STATUS[rows[-1]['status']],
                git_ref=rows[-1]['branch'],
                git_file=rows[-1]['start_script'] or '',
                commit_sha=rows[-1]['commit_sha'],
                startdate=rows[-1]['startdate'],
                enddate=rows[-1]['enddate'],
            )

        for row in rows:
            if DEPLOY_STATUS[row['status']] == Deployment.RUNNING:
                with freeze_time(row['created_at']):
                    deployment.revisions.create(
                        git_ref=row['branch'],
                        git_file=row['start_script'] or '',
                        commit_sha=row['commit_sha'],
                        startdate=row['startdate'],
                        enddate=row['enddate'],
                    )


def import_study(db, row, department, deploy_targets):
    if row['username'] in ['jakob', 'amaddio', 'arc-stele']:
        return

    if Study.objects.filter(title=row['studyname']).exists():
        return

    user, _ = User.objects.get_or_create(username=row['username'], defaults={
        'email': row['email'],
        'is_active': False,
    })

    with freeze_time(row['creationdate']):
        study = Study.objects.create(
            imported_from='ARC',
            title=row['studyname'],
            short_title=slugify(row['study_identifier'] or row['studyname']),
            affiliated_scientists=row['affiliatedscientist'],
            description=(
                row['description']
                .replace('\\r\\n', '\n')
                .replace('\\"', '"')
            ),
            end=row['finisheddate'] or row['datacollectiondone'] or DEFAULT_DATE,
            department=department,
            contact_person=user,
            storage_location=(
                smb.url(row['archivefolderpath'].replace(r'\\', '\\'))
                if row['archivefolderpath'] else None
            ),
        )
    study.owners.add(user)

    for keyword in row['keywords'].replace(', ', '; ').split('; '):
        if keyword.strip():
            study.keyword_set.create(name=keyword.strip())

    if row['expectedcost']:
        import_budgets(db, study, row)

    import_deployments(db, deploy_targets, study, row)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('path')

    def handle(self, *args, **options):
        db = sqlite3.connect(options['path'])
        db.row_factory = sqlite3.Row

        department = Department.objects.get(name='ARC')

        deploy_targets = {
            i: DeployTarget.objects.get(name=name) for i, name in DEPLOY_TARGETS.items()
        }

        rows = db.execute(
            'select '
            'study.id, '
            'studyname, '
            'study_identifier, '
            'affiliatedscientist, '
            'affiliatedra, '  # unused
            'description, '
            'archivefolderpath, '
            'keywords, '
            'creationdate, '
            'startdate, '  # unused
            'finisheddate, '
            'datacollectiondone, '
            'username, '
            'email, '
            'expectedcost, '
            'actualcost, '
            'g3approvalby '
            'from study '
            'join user on user.id = study.userid'
        ).fetchall()

        for row in rows:
            print(row['id'], end='\r')
            import_study(db, row, department, deploy_targets)
