# first export data from castellum:
# manage.py dumpdata studies.Study studies.StudyTag

import json
import sys

from django.core.management.base import BaseCommand
from django.db import models
from django.utils.text import slugify
from freezegun import freeze_time

from study_registration.studies.models import Department
from study_registration.studies.models import Study
from study_registration.users.models import User

DEFAULT_EMAIL = 'info@mpib-berlin.mpg.de'
DEFAULT_DEPARTMENT = 'LIP'
DEFAULT_DATE = '1963-01-01'


def split_seperated_string(s):
    parts = (
        s
        .replace('; ', ', ')
        .replace(' (', ', ').replace(')', '')
        .replace('/', ', ')
        .split(', ')
    )
    return [p.strip() for p in parts if p.strip()]


def find_user(full_name, email=''):
    names = split_seperated_string(full_name)
    if email == DEFAULT_EMAIL or '@example.com' in email:
        email = ''

    if not names and not email:
        return

    q = models.Q()
    if email:
        q |= models.Q(email=email)
    for name in names:
        parts = [p for p in name.split() if p and p != 'Dr.']
        q |= models.Q(first_name__icontains=parts[0], last_name__icontains=parts[-1])

    return User.objects.filter(q).first()


def get_affiliated_scientists(row):
    a = split_seperated_string(row['affiliated_scientists'])
    b = split_seperated_string(row['principal_investigator'])
    merged = []
    for name in a + b:
        if name not in merged:
            merged.append(name)
    return ', '.join(merged)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument('path')

    def handle(self, *args, **options):
        with open(options['path']) as fh:
            rows = json.load(fh)

        default_contact = User.objects.get(email=DEFAULT_EMAIL)
        default_department = Department.objects.get(name=DEFAULT_DEPARTMENT)

        studies = {
            row['pk']: {
                'tags': [],
                **row['fields'],
            }
            for row in rows
            if row['model'] == 'studies.study'
        }

        for row in rows:
            if row['model'] == 'studies.studytag':
                pk = row['fields']['study']
                name = row['fields']['name']
                studies[pk]['tags'].append(name)

        for pk, row in studies.items():
            print(pk, end='\r', file=sys.stderr)

            with freeze_time(row['sessions_start'] or DEFAULT_DATE):
                if Study.objects.filter(title=row['name']).exists():
                    continue

                affiliated_scientists = get_affiliated_scientists(row)
                contact_person = (
                    find_user(row['contact_person'], row['email'])
                    or find_user(affiliated_scientists)
                    or default_contact
                )

                study = Study.objects.create(
                    imported_from='castellum',
                    title=row['name'],
                    short_title=slugify(row['name']),
                    end=row['sessions_end'] or DEFAULT_DATE,
                    description=row['description'].replace('\\r\\n', '\n'),
                    department=contact_person.default_department or default_department,
                    contact_person=contact_person,
                    affiliated_scientists=affiliated_scientists,
                )
                study.owners.add(contact_person)
                for name in row['tags']:
                    study.keyword_set.create(name=name)

                print('{},{}'.format(pk, study.get_absolute_url()))
